/* A simple interface to turtle field     */
/* for JOHO KISO                          */
/* Yoshinori Hayakawa (CITE, Tohoku Univ) */
/* 11-JUL-2013                            */
/* 30-JUN-2017                            */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#define BUFFLEN 512
#define MAX_MSG 128
#define TSLEEP  2000
#define TTLPORT 2222
#define HTTPPORT 2280

#define EVENT_HIT_BY_BULLET   (0x00000001)
#define EVENT_RUN_INTO_TURTLE (0x00000002)
#define EVENT_RUN_INTO_DONUT  (0x00000004)
#define EVENT_RUN_INTO_STONE  (0x00000008)
#define EVENT_RUN_INTO_WALL   (0x00000010)
#define EVENT_FOUND_COIN      (0x00000020)
#define EVENT_DETECTED_BY_FINDER (0x00000040)
#define EVENT_DETECTED_BY_RADAR  (0x00000080)
#define EVENT_GOT_MESSAGE     (0x00000100)

static char ttl_buf_[BUFFLEN] ;
static char ttl_ack_[BUFFLEN] ;
static char *ack_[16] ;
static void (* ttl_hit_by_bullet_handler_)(unsigned int) ;
static void (* ttl_run_into_turtle_handler_)(unsigned int) ;
static void (* ttl_run_into_donut_handler_)(unsigned int) ;
static void (* ttl_run_into_stone_handler_)(unsigned int) ;
static void (* ttl_run_into_wall_handler_)(unsigned int) ;
static void (* ttl_found_coin_handler_)(unsigned int) ;
static void (* ttl_detected_by_finder_handler_)(unsigned int) ;
static void (* ttl_detected_by_radar_handler_)(unsigned int) ;
static void (* ttl_got_message_handler_)(unsigned int) ;

static void ttl_call_event_handler_(unsigned int eventcode, unsigned int timecode) ;

static int socket_=-1 ;
static int team_ =0 ;

/* 
   Notes:
   Students, generally speaking, it is definitely bad idea to describe
   functions in header files. This is only for the purpose to
   reduce your trouble in lessons.
*/

#define DISCONNECT(x) DISCON(x)
static void DISCON()
{
  int i ;
  if (socket_>=0) close(socket_) ;
}

#define CONNECT(x) CON(x)
static int CON(char *hostname)
{
  int i ;
  struct sockaddr_in serveraddr;
  struct hostent *host;
  int rc ;

  if (socket_>=0) {
    fprintf(stderr,"seems you are calling CON() repeatedly.\n") ;
    fprintf(stderr,"use CON() only once in your program.\n") ;
    return -1 ;
  }

  socket_ = socket(AF_INET, SOCK_STREAM, 0);

  memset((char *) &serveraddr, 0, sizeof(serveraddr));
  serveraddr.sin_family = AF_INET;
  serveraddr.sin_port = htons(TTLPORT);

  if (isdigit(hostname[0])) {
    serveraddr.sin_addr.s_addr = inet_addr(hostname);
  }
  else {
    host = gethostbyname(hostname) ;
    if (host == NULL) {
      fprintf(stderr,"no such host: %s\n",hostname) ;
      return -1 ;
    }
    else {
      memcpy(&serveraddr.sin_addr,host->h_addr, host->h_length);
    }
  }

  rc = connect(socket_, (struct sockaddr *)&serveraddr, sizeof(serveraddr));
  if (rc==-1) {
    fprintf(stderr,"cannot connect to %s\n",hostname) ;
    fprintf(stderr,"it seems turtle field is not ready...\n") ;
    exit(0) ;
  }
  ttl_hit_by_bullet_handler_=NULL ;
  ttl_run_into_turtle_handler_=NULL ;
  ttl_run_into_donut_handler_=NULL ;
  ttl_run_into_stone_handler_=NULL ;
  ttl_run_into_wall_handler_=NULL ;
  ttl_found_coin_handler_=NULL ;
  ttl_detected_by_finder_handler_=NULL ;
  ttl_detected_by_radar_handler_=NULL ;
  ttl_got_message_handler_=NULL ;
  return rc ;
}

static int send_ttl_buf_and_receive_ack(void) 
{
  int len,n,k,rc ;
  unsigned int ttl_eventcode ;
  unsigned int ttl_timecode ;

  if (socket_ < 0) return 0 ;
  len = strlen(ttl_buf_) ;
  n = send(socket_, ttl_buf_, len,0) ;
  if (n>0) {
    memset(ttl_buf_, 0, sizeof(ttl_buf_));
    n = recv(socket_, ttl_buf_, sizeof(ttl_buf_),0);
    if (n<=0) {
      fprintf(stderr,"I was killed.. bye !!\n") ;
      exit(0) ;
    }
    ttl_buf_[n]='\0' ;
    strcpy(ttl_ack_, ttl_buf_) ;
    ack_[0] = strtok(ttl_ack_," ") ;
    k=1 ;
    while ((ack_[k]=strtok(NULL," "))!=NULL) {
      k++ ;
    }

    /*  server returns   result(s) [OK|ERR] @ timecode eventcode */
    ttl_eventcode = (unsigned int) atoi(ack_[k-1]) ; /* event code comes always last */
    ttl_timecode = (unsigned int) atoi(ack_[k-2]) ; 
    if ((strncmp(ack_[k-4],"OK",2)==0)) rc=0 ; else rc=-1 ;
    if (ttl_eventcode!=0) ttl_call_event_handler_(ttl_eventcode, ttl_timecode) ;
  } else {
    rc = -1 ;
  }
  usleep(TSLEEP) ;
  return rc ;
}

static void ttl_convert_str2hex(char *str, char *hex)
{
  int cnt=0 ;
  char buf[3] ;
  hex[0]='\0';
  while (*str != '\0' && cnt++<MAX_MSG) {
    sprintf(buf,"%02x",*str) ;
    strcat(hex,buf) ;
    str++ ;
  }
}

static int ttl_char2int(char c)
{
  int r=0 ;
  if (c >= '0' && c<= '9') r = c - '0' ;
  else if (c>='a' && c<='f') r = c - 'a' + 10 ;
  else if (c>='A' && c<='F') r = c - 'A' + 10 ;
  return r ;
}

static void ttl_convert_hex2str(char *hex, char *str)
{
  int cnt,c0,c1 ;
  char buf[2] ;
  str[0]='\0' ;
  while (*hex != '\0' && cnt++<MAX_MSG) {
    c1 = ttl_char2int(*hex) ; hex++ ;
    c0 = ttl_char2int(*hex) ; hex++ ;
    buf[0] = c1*16+c0 ; buf[1]='\0' ;
    strcat(str,buf) ;
  }
}


static int ttl_send_command_(const char *cmd)
{
  strcpy(ttl_buf_,cmd) ;
  return send_ttl_buf_and_receive_ack()  ;
}

int TTL_DISABLE_EVENT(unsigned int eventcode)
{
  sprintf(ttl_buf_,"Y - %u\n",eventcode) ;
  return send_ttl_buf_and_receive_ack()  ;
}

int TTL_ENABLE_EVENT(unsigned int eventcode)
{
  sprintf(ttl_buf_,"Y + %u\n",eventcode) ;
  return  send_ttl_buf_and_receive_ack()  ;
}

static void ttl_call_event_handler_(unsigned int eventcode, unsigned int timecode) {
  int curr_team ;
  char msg_stack[BUFFLEN] ;
  curr_team = team_ ;
  strncpy(msg_stack,ttl_buf_,BUFFLEN) ;
  if (eventcode & EVENT_HIT_BY_BULLET) {
    if (ttl_hit_by_bullet_handler_!=NULL) {
      TTL_DISABLE_EVENT(EVENT_HIT_BY_BULLET) ;
      ttl_hit_by_bullet_handler_(timecode) ;
      TTL_ENABLE_EVENT(EVENT_HIT_BY_BULLET) ;
    }
  }
  if (eventcode & EVENT_RUN_INTO_TURTLE) {
    if (ttl_run_into_turtle_handler_!=NULL) {
      TTL_DISABLE_EVENT(EVENT_RUN_INTO_TURTLE) ;
      ttl_run_into_turtle_handler_(timecode);
      TTL_ENABLE_EVENT(EVENT_RUN_INTO_TURTLE) ;
    }
  }
  if (eventcode & EVENT_RUN_INTO_DONUT) {
    if (ttl_run_into_donut_handler_!=NULL) {
      TTL_DISABLE_EVENT(EVENT_RUN_INTO_DONUT) ;
      ttl_run_into_donut_handler_(timecode);
      TTL_ENABLE_EVENT(EVENT_RUN_INTO_DONUT) ;
    }
  }
  if (eventcode & EVENT_RUN_INTO_STONE) {
    if (ttl_run_into_stone_handler_!=NULL) {
      TTL_DISABLE_EVENT(EVENT_RUN_INTO_STONE) ;
      ttl_run_into_stone_handler_(timecode);
      TTL_ENABLE_EVENT(EVENT_RUN_INTO_STONE) ;
    }
  }
  if (eventcode & EVENT_RUN_INTO_WALL) {
    if (ttl_run_into_wall_handler_!=NULL) {
      TTL_DISABLE_EVENT(EVENT_RUN_INTO_WALL) ;
      ttl_run_into_wall_handler_(timecode);
      TTL_ENABLE_EVENT(EVENT_RUN_INTO_WALL) ;
    }
  }
  if (eventcode & EVENT_FOUND_COIN) {
    if (ttl_found_coin_handler_!=NULL) {
      TTL_DISABLE_EVENT(EVENT_FOUND_COIN) ;
      ttl_found_coin_handler_(timecode);
      TTL_ENABLE_EVENT(EVENT_FOUND_COIN) ;
    }
  }
  if (eventcode & EVENT_DETECTED_BY_FINDER) {
    if (ttl_detected_by_finder_handler_!=NULL) {
      TTL_DISABLE_EVENT(EVENT_DETECTED_BY_FINDER) ;
      ttl_detected_by_finder_handler_(timecode);
      TTL_ENABLE_EVENT(EVENT_DETECTED_BY_FINDER) ;
    }
  }
  if (eventcode & EVENT_DETECTED_BY_RADAR) {
    if (ttl_detected_by_radar_handler_!=NULL) {
      TTL_DISABLE_EVENT(EVENT_DETECTED_BY_RADAR) ;
      ttl_detected_by_radar_handler_(timecode);
      TTL_ENABLE_EVENT(EVENT_DETECTED_BY_RADAR) ;
    }
  }
  if (eventcode & EVENT_GOT_MESSAGE) {
    if (ttl_got_message_handler_!=NULL) {
      TTL_DISABLE_EVENT(EVENT_GOT_MESSAGE) ;
      ttl_got_message_handler_(timecode);
      TTL_ENABLE_EVENT(EVENT_GOT_MESSAGE) ;
    }
  }
  team_ = curr_team ;
  strncpy(ttl_buf_,msg_stack,BUFFLEN) ;
}

void SET_COLLIDE_HIT_BY_BULLET_HANDLER(void (*callbackfunc)(unsigned int)) {
  ttl_hit_by_bullet_handler_ = callbackfunc ;
}

void SET_RUN_INTO_TURTLE_HANDLER(void (*callbackfunc)(unsigned int)) {
  ttl_run_into_turtle_handler_ = callbackfunc ;
}

void SET_RUN_INTO_DONUT_HANDLER(void (*callbackfunc)(unsigned int)) {
  ttl_run_into_donut_handler_ = callbackfunc ;
}

void SET_RUN_INTO_STONE_HANDLER(void (*callbackfunc)(unsigned int)) {
  ttl_run_into_stone_handler_ = callbackfunc ;
}

void SET_RUN_INTO_WALL_HANDLER(void (*callbackfunc)(unsigned int)) {
  ttl_run_into_wall_handler_ = callbackfunc ;

}

void SET_FOUND_COIN_HANDLER(void (*callbackfunc)(unsigned int)) {
  ttl_found_coin_handler_ = callbackfunc ;
}

void SET_DETECTED_BY_FINDER_HANDLER(void (*callbackfunc)(unsigned int)) {
  ttl_detected_by_finder_handler_ = callbackfunc ;
}

void SET_DETECTED_BY_RADAR_HANDLER(void (*callbackfunc)(unsigned int)) {
  ttl_detected_by_radar_handler_ = callbackfunc ;
}

void SET_GOT_MESSAGE_HANDLER(void (*callbackfunc)(unsigned int)) {
  ttl_got_message_handler_ = callbackfunc ;
}

int HOME()
{
  return ttl_send_command_("C H\n") ;
}

int RST()
{
  return ttl_send_command_("C\n") ;
}

int CF()
{
  return ttl_send_command_("C F\n") ;

}

int CLR()
{
  return ttl_send_command_("C A\n") ;
}

int FD(float step)
{
  sprintf(ttl_buf_,"F %f\n",step) ;
  return send_ttl_buf_and_receive_ack()  ;
}

int BK(float step)
{
  sprintf(ttl_buf_,"B %f\n",step) ;
  return send_ttl_buf_and_receive_ack()  ;
}

int JUMP(float x, float y) 
{  
  sprintf(ttl_buf_,"J %f %f\n",x,y) ;
  return send_ttl_buf_and_receive_ack()  ;
}

int RJUMP(float dx, float dy) 
{  
  sprintf(ttl_buf_,"G %f %f\n",dx,dy) ;
  return send_ttl_buf_and_receive_ack()  ;
}

int RT(float angle)
{
  sprintf(ttl_buf_,"T R %f\n",angle) ;
  return send_ttl_buf_and_receive_ack()  ;
}

int LT(float angle)
{
  sprintf(ttl_buf_,"T L %f\n",angle) ;
  return send_ttl_buf_and_receive_ack()  ;
}

int EAST()
{
  return ttl_send_command_("T E\n") ;
}

int WEST()
{
  return ttl_send_command_("T W\n") ;
}

int NORTH()
{
  return ttl_send_command_("T N\n") ;
}

int SOUTH()
{
  return ttl_send_command_("T S\n") ;
}

int LW(float width)
{
  sprintf(ttl_buf_,"W %f\n",width) ;
  return send_ttl_buf_and_receive_ack()  ;
}

int PD()
{
  return ttl_send_command_("P D\n") ;
}

int PU()
{
  return ttl_send_command_("P U\n") ;
}

int FILL()
{
  return ttl_send_command_("L +\n") ;
}

int POINT()
{
  return ttl_send_command_("L P\n") ;
}

int LINE()
{
  return ttl_send_command_("L -\n") ;
}

int BRUSH()
{
  return ttl_send_command_("L B\n") ;
}

int IDLE(float u)
{
  return usleep(u*1000000.0) ;
}

int RF(float range)
{
  sprintf(ttl_buf_,"E %f\n",range) ;
  return send_ttl_buf_and_receive_ack()  ;
}

int SAY(char *something)
{
  int rc,cnt=0 ;
  char hex[MAX_MSG*2+1] ;
  ttl_convert_str2hex(something,hex) ;
  sprintf(ttl_buf_,"S %s\n",hex) ;
  return send_ttl_buf_and_receive_ack()  ;
}

int BCAS(char *msg)
{
  int cnt=0 ;
  char hex[MAX_MSG*2+1] ;
  ttl_convert_str2hex(msg,hex) ;
  sprintf(ttl_buf_,"U %s\n",hex) ;
  return send_ttl_buf_and_receive_ack()  ;
}

int TPRINTF(const char* format, ...)
{
  va_list arg ;
  char buf[BUFFLEN]={0} ;
  va_start(arg,format) ;
  vsprintf(buf,format,arg) ;
  va_end(arg) ;
  return SAY(buf) ;
}

int TSCANF(const char* format, ...)
{
  va_list arg ;
  char buf[BUFFLEN]={0} ;
  va_start(arg,format) ;
  vsscanf(ttl_buf_,format,arg) ;
  va_end(arg) ;
  return 0 ;
}

int ttl_send_packet(char *msg)
{
  if (strlen(msg)>=BUFFLEN) fprintf(stderr,"message is too long\n") ;
  sprintf(ttl_buf_,"%s\n",msg) ;
  return send_ttl_buf_and_receive_ack()  ;
}

int FIRE()
{
  return ttl_send_command_("X\n") ;
}

int ROBOT()
{
  return ttl_send_command_("R\n") ;
}

#define  DOUGHNUT(x) DONUT(x)
int DONUT()
{
  return ttl_send_command_("D\n") ;
}

int BORROWCOIN()
{
  return ttl_send_command_("O B\n") ;
}

int DROPCOIN()
{
  return ttl_send_command_("O S\n") ;
}

int PICKCOIN()
{
  return ttl_send_command_("O R\n") ;
}

int COIN()
{
  return ttl_send_command_("O X\n") ;
}

int COL(float r, float g, float b)
{
  sprintf(ttl_buf_,"K %f %f %f\n",r,g,b) ;
  return send_ttl_buf_and_receive_ack()  ;
}

int BGC(float r, float g, float b)
{
  sprintf(ttl_buf_,"H %f %f %f\n",r,g,b) ;
  return send_ttl_buf_and_receive_ack()  ;
}

int CARD(int i, float x, float y) {
  sprintf(ttl_buf_,"Z %d %f %f\n",i,x,y) ;
  return send_ttl_buf_and_receive_ack()  ;
}

int CLRCARD() {
  sprintf(ttl_buf_,"Z 0\n") ;
  return send_ttl_buf_and_receive_ack()  ;
}

int BMODE()
{
  return ttl_send_command_("! +\n") ;
}

int GMODE()
{
  return ttl_send_command_("! -\n") ; 
}

int NM(char *s)
{
  int i ;
  char hex[MAX_MSG*2+1] ;
  ttl_convert_str2hex(s,hex) ;
  sprintf(ttl_buf_,"N %s\n",hex) ;
  return send_ttl_buf_and_receive_ack()  ;
}

int TM(int team)
{
  sprintf(ttl_buf_,"I %d\n",team) ;
  return send_ttl_buf_and_receive_ack()  ;
}


int Q_ID(int *tid)
{
  int rc,t ;
  char buf[32] ;
  sprintf(ttl_buf_,"Q M\n") ;
  rc = send_ttl_buf_and_receive_ack()  ;
  if (rc>=0) {
    sscanf(ttl_buf_,"%d %s",&t,buf) ;
    *tid = t ;
  }
  else *tid = -1 ;
  return rc ;
}

int Q_POS(float *x, float *y)
{
  int rc ;
  float rx,ry ;
  char buf[32] ;
  sprintf(ttl_buf_,"Q P\n") ;
  rc = send_ttl_buf_and_receive_ack()  ;
  if (rc>=0) {
    sscanf(ttl_buf_,"%f %f %s",&rx,&ry,buf) ;
    *x = rx ; *y = ry ;
  }
  else { 
    *x= 0 ; *y = 0 ; 
  }
  return rc ;
}

#define Q_HEADING(x) Q_DIR(x)
int Q_DIR(float *dir)
{
  int rc ;
  float rd ;
  char buf[32] ;
  sprintf(ttl_buf_,"Q D\n") ;
  rc = send_ttl_buf_and_receive_ack()  ;
  if (rc>=0) {
    sscanf(ttl_buf_,"%f %s",&rd,buf) ;
    *dir = rd ;
  }
  else { 
    *dir= 0 ; 
  }
  return rc ;
}

int Q_ALIVE(int *yn)
{
  int rc,stat=0 ;
  char buf[32] ;
  if (socket_ > 0) {
    int len,n ;
    sprintf(ttl_buf_,"Q A\n") ;
    len = strlen(ttl_buf_) ;
    n = write(socket_, ttl_buf_, len) ;
    if (n>0) {
      memset(ttl_buf_, 0, sizeof(ttl_buf_));
      n = read(socket_, ttl_buf_, sizeof(ttl_buf_));
      if (n<=0) stat=0 ;
      else stat=1 ;
    }
  }
  *yn = stat ;
  return rc ;
}

int Q_SCORE(int *score)
{
  int rc,s ;
  char buf[32] ;
  sprintf(ttl_buf_,"Q S\n") ;
  rc = send_ttl_buf_and_receive_ack()  ;
  if (rc>=0) {
    sscanf(ttl_buf_,"%d %s",&s,buf) ;
    *score = s ;
  }
  else { 
    *score= 0 ; 
  }
  return rc ;
}

int Q_NT(int *nt)
{
  int rc,na ;
  char buf[32] ;
  sprintf(ttl_buf_,"Q N\n") ;
  rc = send_ttl_buf_and_receive_ack()  ;
  if (rc>=0) {
    sscanf(ttl_buf_,"%d %s",&na,buf) ;
    *nt = na ;
  }
  else { 
    *nt = 0 ; 
  }
  return rc ;
}

int Q_COIN(int *nc)
{
  int rc,n ;
  char buf[32] ;
  sprintf(ttl_buf_,"Q C\n") ;
  rc = send_ttl_buf_and_receive_ack()  ;
  if (rc>=0) {
    sscanf(ttl_buf_,"%d %s",&n,buf) ;
    *nc = n ;
  }
  else { 
    *nc = 0 ; 
  }
  return rc ;
}

int Q_MYCOIN(int *nmyc)
{
  int rc,n ;
  char buf[32] ;
  sprintf(ttl_buf_,"Q O\n") ;
  rc = send_ttl_buf_and_receive_ack()  ;
  if (rc>=0) {
    sscanf(ttl_buf_,"%d %s",&n,buf) ;
    *nmyc = n ;
  }
  else { 
    *nmyc = 0 ; 
  }
  return rc ;
}

int Q_RADAR(float *dir)
{
  int rc,team ;
  float angle ;
  char buf[32] ;
  sprintf(ttl_buf_,"Q R\n") ;
  rc = send_ttl_buf_and_receive_ack()  ;
  if (rc>=0) {
    sscanf(ttl_buf_,"%f %d %s",&angle,&team,buf) ;
    *dir = angle ;
    team_ = team ;
  }
  else { 
    *dir= 0 ; 
    team_ = 0 ;
  }
  return rc ;
}

int Q_RADAR_N(float *dir, int nn)
{
  int rc,team ;
  float angle ;
  char buf[32] ;
  sprintf(ttl_buf_,"Q R %d\n",nn) ;
  rc = send_ttl_buf_and_receive_ack()  ;
  if (rc>=0) {
    sscanf(ttl_buf_,"%f %d %s",&angle,&team,buf) ;
    *dir = angle ;
    team_ = team ;
  }
  else { 
    *dir= 0 ; 
    team_ = 0 ;
  }
  return rc ;
}

int Q_FINDER(int *yn)
{
  int rc, res1, res2 ;
  char buf[32] ;
  sprintf(ttl_buf_,"Q F\n") ;
  rc = send_ttl_buf_and_receive_ack()  ;
  if (rc>=0) {
    sscanf(ttl_buf_,"%d %d %s",&res1,&res2,buf) ;
    *yn = res1 ;
    team_ = res2 ;
  }
  else { 
    *yn = 0 ;
    team_ = 0 ;
  }
  return rc ;
}


int Q_SONAR(int *left, int *front, int *right)
{
  int rc,res1,res2,res3 ;
  char buf[32] ;
  sprintf(ttl_buf_,"Q W\n") ;
  rc = send_ttl_buf_and_receive_ack()  ;
  if (rc>=0) {
    sscanf(ttl_buf_,"%d %d %d %s",&res1,&res2,&res3,buf) ;
    *left = res1 ;
    *front = res2 ;
    *right = res3 ;
  }
  else {
    *left = 0 ;
    *front = 0 ;
    *right = 0 ;
  }
  return rc ;
}

int Q_TM(int *team)
{
  *team = team_ ;
  return 0 ;
}

int Q_BCAS(int *etime, char *msg)
{
  int rc, res ;
  char buf[MAX_MSG*2+1],rest[BUFFLEN] ;
  sprintf(ttl_buf_,"Q B\n") ;
  rc = send_ttl_buf_and_receive_ack()  ;
  if (rc>=0) {
    sscanf(ttl_buf_,"%d%s%[^\n]",&res,buf,rest) ;
    ttl_convert_hex2str(buf,msg) ;
    *etime = res ;
  }
  return rc ;
}

int Q_TIME(int *tim)
{
  int rc, res ;
  char buf[32] ;
  sprintf(ttl_buf_,"Q T\n") ;
  rc = send_ttl_buf_and_receive_ack()  ;
  if (rc>=0) {
    sscanf(ttl_buf_,"%d %s",&res,buf) ;
    *tim = res ;
  }
  else { 
    *tim = 0 ;
  }
  return rc ;
}

int Q_STAT_CARD(int c, int *stat)
{
  int rc ;
  char buf[16] ;
  sprintf(ttl_buf_,"Q Z S %d\n",c) ;
  rc = send_ttl_buf_and_receive_ack()  ;
  if (rc>=0) {
    sscanf(ttl_buf_,"%d %s",stat,buf) ;
  } else {
    *stat = -1 ;
  }
  return rc ;
}

int Q_POS_CARD(int c, float *x, float *y)
{
  int rc ;
  char buf[16] ;
  sprintf(ttl_buf_,"Q Z P %d\n",c) ;
  rc = send_ttl_buf_and_receive_ack()  ;
  if (rc>=0) {
    sscanf(ttl_buf_,"%f %f %s",x,y,buf) ;
  }

  return rc ;
}

